#include "TYEGFiles.h"
#include "core/core.h"
#include "shape/ShapeManager.h"
#include "shape/TShapeFactory.h"
#include "shape/TShape.h"
#include "tools/DefaultTool/TDefaultTool.h"
#include "widgets/canvas/TWView.h"
#include "widgets/canvas/TConversion.h"
#include "shape/TSelectionShape.h"
#include "shape/TLinesShape.h"

#include <QFile>
#include <QMessageBox>
#include <QDebug>

namespace graphed {
namespace out {

TYEGFiles::TYEGFiles()
{
}

void TYEGFiles::Save(const QString &FileName)
{
    QFile File(FileName);
    if (!File.open(QIODevice::WriteOnly)) {
        qDebug() << "Unable to open file for write" + FileName;
        return;
    }
    File.resize(0); QDataStream DataStream(&File);
    const TShapeList &Shapes = INSTANCE(ShapeManager)->ShapeList();
    foreach (TShapeSP Shape, Shapes)
    {
        if (Shape->objectName() == GraphEd::ResultShapeName() || Shape->objectName() == GraphEd::BackgroundShapeName()) continue;
        Shape->SaveToFile(DataStream);
    }
    File.close();
}

TShapeList TYEGFiles::Load(const QString &FileName)
{
    TShapeList Result;

    QFile File(FileName);
    if (!File.open(QIODevice::ReadOnly)) {
        qDebug() << "Unable to open file for read" + FileName;
        return Result;
    }

    if (FileName.right(4) == ".dxf" || FileName.right(4) == ".DXF" ) {
        QTextStream DataStream(&File);
        TShapeSP Shape;
        int depth = 0;
        quint32 Id = 0;
        QString tagName;
        static qreal scale = 500;
        bool spline_ignore_fit = false;
        bool ellipse_ignore_rscale = false;

        tagName = DataStream.readLine();
        if (tagName.indexOf("AutoCAD Binary DXF") == 0) {
            QMessageBox mb(QMessageBox::Warning,
                           QObject::tr("DXF Import error"),
                           QObject::tr("Binary DXF is not supported.\nContact developer for further assistance."),
                           QMessageBox::Close);
            mb.exec();
            File.close();
            return Result;
        }

        INSTANCE(ShapeManager)->Selection()->DeSelectAll();

        bool block_flag = false;

        while (!DataStream.atEnd()) {
            if (Id == 0) {
                DataStream >> tagName;

                if (depth > 0) {
                    depth--;
                    if (Shape) {
                        if (!block_flag) {
                            Result << Shape;
                            INSTANCE(ShapeManager)->Selection()->Select(Shape);
                        }

                        Shape.clear();
                    }
                }

                if (tagName == "EOF") {
                    break;
                } else if (tagName == "BLOCK") {
                    block_flag = true;
                } else if (tagName == "ENDBLK") {
                    block_flag = false;
                } else if (tagName == "POINT") {
                    Shape = INSTANCE(TShapeFactory)->Create("TPointShape");
                    depth++;
                } else if (tagName == "LINE" ||
                           //tagName == "MLINE" || //untested!!!
                           tagName == "XLINE" || //untested
                           tagName == "SPLINE" ||
                           tagName == "LWPOLYLINE") {
                    Shape = INSTANCE(TShapeFactory)->Create("TLinesShape");
                    depth++;
                } else if (tagName == "CIRCLE") {
                    Shape = INSTANCE(TShapeFactory)->Create("TEllipseShape");
                    depth++;
                } else if (tagName == "ARC" ||
                           tagName == "ELLIPSE") {
                    Shape = INSTANCE(TShapeFactory)->Create("TEllipseCurveShape");
                    depth++;
                }

                if (tagName == "ELLIPSE")
                    ellipse_ignore_rscale = true;
                else
                    ellipse_ignore_rscale = false;

                if (tagName == "SPLINE")
                    spline_ignore_fit = true;
                else
                    spline_ignore_fit = false;

            } else if (Id == 10 || Id == 11 ||
                       Id == 20 || Id == 21) {
                qreal pos;
                DataStream >> pos;

                pos *= scale;
                if (Id == 20 || Id == 21)
                    pos = -pos;

                if (!spline_ignore_fit
                        || (Id != 11 && Id != 21))
                {
                    if (Shape)
                        Shape->SetFloatParameter(Id, pos);
                }
            } else if (Id == 40) {
                qreal radius;
                DataStream >> radius;

                if (!ellipse_ignore_rscale)
                    radius *= scale;

                if (Shape)
                    Shape->SetFloatParameter(Id, radius);
            } else if (Id == 41 || Id == 42 ||
                       Id == 50 || Id == 51) {
                qreal angle;
                DataStream >> angle;

                if (Shape)
                    Shape->SetFloatParameter(Id, angle);
            } else if (Id == 70) {
                quint32 flags;
                DataStream >> flags;

                if (Shape)
                    Shape->SetIntParameter(Id, flags);
            } else if (Id == 90 || Id == 73) {
                quint32 pnum;
                DataStream >> pnum;

                if (Shape)
                    Shape->SetIntParameter(Id, pnum);
            } else {
                DataStream.readLine();//skip lineend from id line
                DataStream.readLine();
            }

            DataStream >> Id;
        }

        INSTANCE(ShapeManager)->Selection()->InvalidateCache();
        TConversionBase *Conversion = INSTANCE(TWView)->Conversion();
        QRectF rect = Conversion->DocumentRect();

        qreal k = 1;

        if (INSTANCE(ShapeManager)->Selection()->BoundingRect().width() != 0)
            k = INSTANCE(ShapeManager)->Selection()->BoundingRect().height() /
                    INSTANCE(ShapeManager)->Selection()->BoundingRect().width();

        qreal width = rect.width()*1e6*0.95;
        qreal height = rect.height()*1e6*0.95;

        if (k > 1)
            width = height / k;
        else
            height = width * k;

        SetSelectionParams(rect.width()*1e6/2,
                           rect.height()*1e6/2,
                           width,
                           height);

        INSTANCE(ShapeManager)->Selection()->DeSelectAll();
        INSTANCE(ShapeManager)->Selection()->InvalidateCache();
    } else {
        QDataStream DataStream(&File);
        while (!File.atEnd())
        {
            QString ShapeName;
            DataStream >> ShapeName;
            if (TShapeSP Shape = INSTANCE(TShapeFactory)->Create(ShapeName))
            {
                Shape->LoadFromFile(DataStream);
                Result << Shape;
            }
        }
    }

    File.close();

    return Result;
}

}
}
