#include "ShapeManager.h"
#include "shape/TSelectionShape.h"
#include "tools/ToolManager.h"
#include <QDebug>

namespace graphed {

class ShapeManagerPrivate
   {
   public:
      explicit ShapeManagerPrivate();
      virtual ~ShapeManagerPrivate();
      TShapeList AllShapes;
      TSelectionShapeSP Selection;
      TCanvasQP Canvas;
   };

ShapeManagerPrivate::ShapeManagerPrivate()
   {
      Selection = TSelectionShapeSP(new TSelectionShape);
      Canvas = INSTANCE(TCanvas);
   }
ShapeManagerPrivate::~ShapeManagerPrivate()
   {
      //delete Selection; Selection = NULL;
   }


ShapeManager *ShapeManager::Self = NULL;
bool ShapeManager::IAmBeingCreatedNow = false;
ShapeManager::ShapeManager(): d_ptr(new ShapeManagerPrivate) { }
ShapeManager::~ShapeManager() { delete d_ptr; }
ShapeManager *ShapeManager::Instance(const int Line, const char *File)
   {
      if (IAmBeingCreatedNow)
         { qDebug() << "ShapeManager::Instance() was called from " << File << " and line #" << Line << "! Returning NULL..."; return NULL; }
      if (!Self)
         {
            IAmBeingCreatedNow = true;
            Self = new ShapeManager;
            IAmBeingCreatedNow = false;
         }
      return Self;
   }
void ShapeManager::Destroy()  { delete Self; Self = 0;  }

void ShapeManager::AddShape(TShapeSP s)
   {
      Q_D(ShapeManager);
      d->AllShapes.push_back(s);
      emit ShapeChanged(s);
   }
void ShapeManager::paint(QPainter *p, const TConversionBase *Convert) const
   {
      Q_D(const ShapeManager);

      TShapeList Sorted = d->AllShapes;
      qSort(Sorted.begin(), Sorted.end(), TShape::CompareShapeByZIndex);

      foreach (TShapeSP Shape, Sorted)
         {
            p->save();
            TShape::ApplyConversion(p, Convert);    //VVVVVVVVVVVVV - Do not loose painter translation from TCanvas::update()
            p->setTransform(Shape->Transformation() * p->transform());
            Shape->paint(p, Convert);
            p->restore();
         }
   }
TShapeSP ShapeManager::ShapeAt(const QPointF &p, GraphEd::TShapeSelection SelMode) const
   {
      Q_D(const ShapeManager);
      TShapeList Sorted;
      foreach (TShapeSP Shape, d->AllShapes) if (Shape->IsIn(p)) Sorted << Shape;
      qSort(Sorted.begin(), Sorted.end(), TShape::CompareShapeByZIndex);
      for (int i = Sorted.count() - 1; i >= 0; i--)
         {
            const TShapeSP &Shape = Sorted[i];
            //qDebug() << Shape->objectName() << Shape->ZIndex() << d->Selection->IsSelected(Shape);
            switch(SelMode)
               {
                  case GraphEd::ShapeOnTop:
                     if (Shape->Selectable()) return Shape;
                     break;
                  case GraphEd::Selected :
                     if ( d->Selection->IsSelected(Shape)) return Shape;
                     break;
                  case GraphEd::Unselected:
                     if (!d->Selection->IsSelected(Shape)) return Shape;
                     break;
               }
         }
      return TShapeSP();
   }
const TSelectionShapeSP ShapeManager::Selection() const
   {
      Q_D(const ShapeManager);
      return d->Selection;
   }
TSelectionShapeSP ShapeManager::Selection()
   {
      Q_D(ShapeManager);
      return d->Selection;
   }
/*void ShapeManager::update()
   {
      Q_D(ShapeManager);
      d->Canvas->update();
      //if (selectionHandles && d->selection->isSelected(shape)) if (d->Canvas->toolProxy()) d->Canvas->toolProxy()->repaintDecorations();
   }*/
TShapeList ShapeManager::FindAll(const QString &ObjName) const
   {
      TShapeList Result;
      Q_D(const ShapeManager);
      foreach (TShapeSP Shape, d->AllShapes)
         if (Shape->objectName() == ObjName) Result << Shape;
      return Result;
   }

void ShapeManager::RemoveAll(const QString &ObjName)
   {
      Q_D(ShapeManager);
      const TShapeList::size_type OldCount = d->AllShapes.count();
      foreach (TShapeSP Shape, d->AllShapes)
         if (Shape->objectName() == ObjName) d->AllShapes.removeAll(Shape);
      if (OldCount != d->AllShapes.count()) emit ShapeChanged(TShapeSP());
   }
void ShapeManager::RemoveAllExcept(const QStringList &Names)
   {
      Q_D(ShapeManager);
      const TShapeList::size_type OldCount = d->AllShapes.count();
      TShapeList Except;
      foreach (TShapeSP Shape, d->AllShapes)
         if (Names.contains(Shape->objectName())) Except.push_back(Shape);
      d->AllShapes.clear();
      d->AllShapes.append(Except);
      if (OldCount != d->AllShapes.count()) emit ShapeChanged(TShapeSP());
   }
void ShapeManager::RemoveAllExcept(const QString &Name)
   {
      RemoveAllExcept(QStringList(Name));
   }
void ShapeManager::Remove(const TShapeList &Cont)
   {
      Q_D(ShapeManager);
      for (TShapeList::const_iterator It = Cont.begin(); It != Cont.end();  d->AllShapes.removeAll(*(It++)));
   }
QRectF ShapeManager::BoundingRect() const
   {
      QRectF Result;
      Q_D(const ShapeManager);
      foreach (TShapeSP Shape, d->AllShapes) Result |= Shape->MappedBoundingRect();
      return Result;
   }
void ShapeManager::NotifyShapeChanged(TShape *Sh)
   {
      Q_D(ShapeManager);
      foreach (TShapeSP Shape, d->AllShapes)
         if (Shape.data() == Sh) emit ShapeChanged(Shape);
   }
TShapeList::size_type ShapeManager::count() const
   {
      Q_D(const ShapeManager);
      return d->AllShapes.size();
   }
const TShapeList &ShapeManager::ShapeList() const
   {
      Q_D(const ShapeManager);
      return d->AllShapes;
   }
TShapeList &ShapeManager::ShapeList()
   {
      Q_D(ShapeManager);
      return d->AllShapes;
   }
int ShapeManager::MaxZIndex() const
   {
      Q_D(const ShapeManager);
      int Z = -1;
      foreach (TShapeSP Shape, d->AllShapes)
         Z = qMax(Z, Shape->ZIndex());
      return Z;
   }
}
