#include "TImageShape.h"
#include "TImageShape_p.h"
#include "core/core.h"
#include "out/TConvertor.h"
#include "out/TResults.h"
#include "widgets/canvas/TWView.h"
#include "widgets/canvas/TConversion.h"

namespace graphed {

TImageShapePrivate::TPixelPartInfoList TImageShapePrivate::PixelPartInfoList(const QRectF &Rect) const
{
    TPixelPartInfoList Result;
    const int MinX = Rect.left() >= 0.   ? ToZero(Rect.left())     : FromZero(Rect.left());
    const int MinY = Rect.top() >= 0.    ? ToZero(Rect.top())      : FromZero(Rect.top());
    const int MaxX = Rect.right() >= 0.  ? FromZero(Rect.right())  : ToZero(Rect.right());
    const int MaxY = Rect.bottom() >= 0. ? FromZero(Rect.bottom()) : ToZero(Rect.bottom());

    for (int XPos = MinX; XPos < MaxX; XPos++)
        for (int YPos = MinY; YPos < MaxY; YPos++)
        {
            qreal Width = 1.;
            qreal Height = 1.;
            if (IsBetween(XPos, XPos + 1, Rect.left()))       Width = qMin(static_cast<qreal>(XPos + 1), Rect.right()) - Rect.left();
            else if (IsBetween(XPos, XPos + 1, Rect.right())) Width = Rect.right() - qMax(static_cast<qreal>(XPos), Rect.left()) ;

            if (IsBetween(YPos, YPos + 1, Rect.top()))         Height = qMin(static_cast<qreal>(YPos + 1), Rect.bottom()) - Rect.top();
            else if (IsBetween(YPos, YPos + 1, Rect.bottom())) Height = Rect.bottom() - qMax(static_cast<qreal>(YPos), Rect.top()) ;

            Result.push_back(TPixelPartInfo(QPoint(XPos, YPos), Width*Height));
        }
    return Result;
}

TImageShape::TImageShape(): TDataShape<out::TVtData>(*new TImageShapePrivate)
{
    setObjectName(ShapeName());
    MetaData1().first = out::TVoltColourConvertor::MinVoltage;
}
void TImageShape::paint(QPainter *p, const TConversionBase * /*Convert*/) const
{
    if (isNull()) return;
    p->drawImage(0, 0, *this);
}
QRectF TImageShape::BoundingRect() const
{
    return QRectF(0, 0, size().width(), size().height());
}
bool TImageShape::IsIn(const QPointF &PF) const
{
    Q_D(const TImageShape);
    return TShape::IsIn(d->Transformation.inverted().map(PF));
}
out::TResults *TImageShape::ToPointList() const
{
    Q_D(const TImageShape);

    TConversionBase *Conversion = INSTANCE(TWView)->Conversion();
    const QRect AllInfoRectInResCoord = Conversion->VirginVRToResolutionF(d->Transformation.mapRect(QRectF(0, 0, width(), height()))).toAlignedRect();
    TImageShapePrivate::TPixelInfoList AllInfo(AllInfoRectInResCoord.width() * AllInfoRectInResCoord.height());

    const unsigned int TotalSteps = height() * width();
    unsigned int CurrentStep = 0;

    for (int YPos = 0; YPos < height(); YPos++)
        for (int XPos = 0; XPos < width(); XPos++)
        {
            SetProgress((CurrentStep++ * 100) / TotalSteps);
            const QRgb Colour = pixel(XPos, YPos);
            const QRectF ImgRect(XPos, YPos, 1, 1);
            const QRectF DocRect = d->Transformation.mapRect(ImgRect);
            const QRectF ResolutionRect = Conversion->VirginVRToResolutionF(DocRect);
            TImageShapePrivate::TPixelPartInfoList PixelPartInfoList = d->PixelPartInfoList(ResolutionRect);
            foreach (TPixelPartInfo PartInfo, PixelPartInfoList)
            {
                TImageShapePrivate::TPixelInfoList::iterator It = AllInfo.begin() + PartInfo.Point().x() - AllInfoRectInResCoord.x() + (PartInfo.Point().y() - AllInfoRectInResCoord.y()) * AllInfoRectInResCoord.width();
                if (It >= AllInfo.end() || It < AllInfo.begin()) continue;
                It->Voltage += PartInfo.Part() * out::TVoltColourConvertor::Colour2Voltage(Colour, 0, MetaData1().first) * qAlpha(Colour) / 255. ;
                It->SetPoint(PartInfo.Point());
            }
        }
    STLPointFVector PList;
    STLRealVector VList;
    foreach (TPixelInfo PInfo, AllInfo)
    {
        PList.push_back(Conversion->ResolutionToPhysical(PInfo.Point()));
        VList.push_back(PInfo.Voltage);
    }

    return new out::TVarVoltage(PList, MetaData1().second, VList, Speed(), MetaData1().first);
}
void TImageShape::SaveToFile(QDataStream &Stream) const
{
    Stream << ShapeName();
    SaveTransform(Stream);
    Stream << *dynamic_cast<const QImage *>(this);
}
void TImageShape::LoadFromFile(QDataStream &Stream)
{
    LoadTransform(Stream);
    Stream >> *dynamic_cast<QImage *>(this);
}


}
