#include "TLinesShape.h"
#include "TLinesShape_p.h"
#include "widgets/canvas/TWView.h"
#include "widgets/canvas/TConversion.h"
#include <math.h>

namespace graphed {

TSettableVar_RF TLinesShapePrivate::UpdateBoundingRect() const
{
    BoundingRect = QRectF();
    if (!Points.size()) return BoundingRect;

    BoundingRect.Var.setRight(Points.front().x());
    BoundingRect.Var.setLeft(Points.front().x());
    BoundingRect.Var.setTop(Points.front().y());
    BoundingRect.Var.setBottom(Points.front().y());

    for (TPointFList::const_iterator It = Points.begin() + 1; It != Points.end(); It++)
        UpdateBoundingRect(BoundingRect.Var, *It);

    if (BoundingRect.Var.width() == 0) {
        qreal dx;
        qreal dy;
        Transformation.inverted().map(1, 0, &dx, &dy);
        dx -= Transformation.inverted().dx();
        dy -= Transformation.inverted().dy();

        dx = sqrt(dx*dx + dy*dy);

        BoundingRect.Var.setRight(BoundingRect.Var.right() + dx);
        BoundingRect.Var.setLeft(BoundingRect.Var.left() - dx);
    }

    if (BoundingRect.Var.height() == 0) {
        qreal dx;
        qreal dy;
        Transformation.inverted().map(0, 1, &dx, &dy);
        dx -= Transformation.inverted().dx();
        dy -= Transformation.inverted().dy();

        dy = sqrt(dx*dx + dy*dy);

        BoundingRect.Var.setTop(BoundingRect.Var.top() - dy);
        BoundingRect.Var.setBottom(BoundingRect.Var.bottom() + dy);
    }

    return BoundingRect;
}
void TLinesShapePrivate::UpdateBoundingRect(QRectF &r, const QPointF &PF) const
{
    if (PF.x() < r.left()) r.setLeft(PF.x());
    if (PF.x() > r.right()) r.setRight(PF.x());
    if (PF.y() > r.bottom()) r.setBottom(PF.y());
    if (PF.y() < r.top()) r.setTop(PF.y());
}

//======================================================================================================================

TLinesShape::TLinesShape(): TDataShape<out::TVtData>(*new TLinesShapePrivate)
{
    setObjectName(ShapeName());
    Q_D(TLinesShape);
    d->Transformation = QTransform();
    d->closed_spline = false;
    d->points_count = 0;

    d->__x_set = false;
    d->__y_set = false;
}

void TLinesShape::SetFloatParameter(int id, qreal val)
{
    Q_D(TLinesShape);

    if (id == 10 || id == 11) {
        d->__x_val = val;
        d->__x_set = true;
    } else if (id == 20 || id == 21) {
        d->__y_val = val;
        d->__y_set = true;
    }

    if (d->__x_set && d->__y_set) {
        d->__x_set = false;
        d->__y_set = false;
        AddPoint(QPointF(d->__x_val, d->__y_val));
    }

    if (d->points_count != 0 && d->closed_spline &&
            (d->Points.size() == d->points_count)) {
        AddPoint(d->Points.first());
    }
}

void TLinesShape::SetIntParameter(int id, quint32 val)
{
    Q_D(TLinesShape);

    if (id == 70) {
        if (val & 0x01)
            d->closed_spline = true;
    } else if (id == 90 || id == 73) {
        d->points_count = val;
    }
}

void TLinesShape::AddPoint(const QPointF &p)
{
    Q_D(TLinesShape);
    d->Points.push_back(p);
    d->UpdateBoundingRect();
}
void TLinesShape::SetFreePoint(const QPointF &p)
{
    Q_D(TLinesShape);
    d->FreePoint = p;
    d->UpdateBoundingRect();
}
void TLinesShape::UnsetFreePoint()
{
    Q_D(TLinesShape);
    d->FreePoint.ValueIsSet = false;
    d->BoundingRect.ValueIsSet = false;
}
void TLinesShape::paint(QPainter *p, const TConversionBase * /*Convert*/) const
{
    Q_D(const TLinesShape);
    p->setPen(d->Pen);
    //p->drawPath(PainterPath());
    for (TPointFList::const_iterator It = d->Points.begin(); It != d->Points.end() - 1; It++)
        p->drawLine(*It, It[1]);
    if (d->FreePoint.IsValueSet())
        p->drawLine(d->Points.back(), d->FreePoint);
}
QRectF TLinesShape::BoundingRect() const
{
    Q_D(const TLinesShape);
    //if (d->BoundingRect.IsValueSet()) return d->BoundingRect;
    //else
    return d->UpdateBoundingRect();
}
bool TLinesShape::IsIn(const QPointF &PF) const
{
    Q_D(const TLinesShape);
    return TShape::IsIn(d->Transformation.inverted().map(PF));
}
TPointFList::size_type TLinesShape::PointsCount() const
{
    Q_D(const TLinesShape);
    return d->Points.size();
}
out::TResults *TLinesShape::ToPointList() const
{
    Q_D(const TLinesShape);
    STLPointFVector Result;
    TConversionBase *Conversion = INSTANCE(TWView)->Conversion();

    for (TPointFList::const_iterator It = d->Points.begin(); It != d->Points.end() - 1; It++)
    {
        const QPoint p1 = Conversion->VirginVRToResolution(d->Transformation.map(*It));
        const QPoint p2 = Conversion->VirginVRToResolution(d->Transformation.map(It[1]));
        const STLPointVector &Line = out::Line(p1, p2);

        for (STLPointVector::const_iterator LineIt = Line.begin(); LineIt != Line.end(); LineIt++)
            Result.push_back(Conversion->ResolutionToPhysical(*LineIt));
    }

    return new out::TConstVoltageRes(Result, MetaData1(), Speed());
}
void TLinesShape::SaveToFile(QDataStream &Stream) const
{
    Q_D(const TLinesShape);
    Stream << ShapeName();
    SaveTransform(Stream);
    Stream << d->Points;
}
void TLinesShape::LoadFromFile(QDataStream &Stream)
{
    Q_D(TLinesShape);
    LoadTransform(Stream);
    Stream >> d->Points;
    d->UpdateBoundingRect();
}


}
