#include "TShape.h"
#include "TShape_p.h"
#include "widgets/canvas/TConversionBase.h"
#include "shape/ShapeManager.h"
#include "core/core.h"


namespace graphed {

TShapePrivate::TShapePrivate(): ZIndex(0), Selectable(true), GeometryProtected(false) { }

TShape::TShape(TShapePrivate &dd): d_ptr(&dd)   {  }
TShape::TShape(): d_ptr(new TShapePrivate)   {  }
TShape::~TShape() { delete d_ptr; }
void TShape::update() const { }
void TShape::SetPosition(const QPointF &Pos)
   {
      QPointF CurrentPos = Position();
      if (CurrentPos == Pos) return;
      Q_D(TShape);
      QTransform TranslateMatrix;
      TranslateMatrix.translate(Pos.x() - CurrentPos.x(), Pos.y() - CurrentPos.y());
      d->Transformation = d->Transformation * TranslateMatrix;
      NotifyChanged();
   }
QPointF TShape::Position() const
   {
      Q_D(const TShape);
      return d->Transformation.map(BoundingRect().center());
   }
QTransform TShape::Transformation() const
   {
      Q_D(const TShape);
      return d->Transformation;
   }
bool TShape::IsIn(const QPointF &PF) const
   {
      return BoundingRect().contains(PF);
   }
QPointF TShape::GlobalPosition(const QPointF &Pos) const
   {
      Q_D(const TShape);
      return d->Transformation.map(Pos);
   }
QPointF TShape::GlobalPosition() const
   {
      return GlobalPosition(Position());
   }
QRectF TShape::MappedBoundingRect() const
   {
      Q_D(const TShape);
      return d->Transformation.mapRect(BoundingRect());
   }
void TShape::SetTransformation(const QTransform &t)
   {
      Q_D(TShape);
      d->Transformation = t;
      NotifyChanged();
   }
void TShape::ApplyConversion(QPainter *p, const TConversionBase *Convert)
   {
      if (!p || !Convert) return;
      p->scale(Convert->Zoom(), Convert->Zoom());
   }
void TShape::NotifyChanged()
   {
      INSTANCE(ShapeManager)->NotifyShapeChanged(this);
   }
void TShape::SaveTransform(QDataStream &Stream) const
   {
      Q_D(const TShape);
      Stream << d->Transformation.m11() << d->Transformation.m12() << d->Transformation.m13() <<
                d->Transformation.m21() << d->Transformation.m22() << d->Transformation.m23() <<
                d->Transformation.m31() << d->Transformation.m32() << d->Transformation.m33();
   }
void TShape::LoadTransform(QDataStream &Stream)
   {
      Q_D(TShape);
      qreal m11, m12, m13, m21, m22, m23, m31, m32, m33;
      Stream >> m11 >> m12 >> m13 >>
                m21 >> m22 >> m23 >>
                m31 >> m32 >> m33;
      d->Transformation = QTransform(m11, m12, m13, m21, m22, m23, m31, m32, m33);
   }
bool TShape::Selectable() const
   {
      Q_D(const TShape);
      return d->Selectable;
   }
void TShape::SetSelectable(const bool S)
   {
      Q_D(TShape);
      d->Selectable = S;
   }
bool TShape::GeometryProtected() const
   {
      Q_D(const TShape);
      return d->GeometryProtected;
   }
void TShape::SetGeometryProtected(const bool GP)
   {
      Q_D(TShape);
      d->GeometryProtected = GP;
   }
bool TShape::CompareShapeByZIndex(TShapeSP S1, TShapeSP S2)
   {
      return S1->ZIndex() < S2->ZIndex();
   }
void TShape::SetZIndex(const int Z)
   {
      Q_D(TShape);
      d->ZIndex = Z;
      NotifyChanged();
   }
int TShape::ZIndex() const
   {
      Q_D(const TShape);
      return d->ZIndex;
   }

/*
void TShape::SetScale(qreal sx, qreal sy)
   {
      Q_D(TShape);
      QPointF Pos = Position();
      d->Transformation.translate(Pos.x(), Pos.y());
      d->Transformation.scale(sx, sy);
      d->Transformation.translate(-Pos.x(), -Pos.y());

      emit ShapeChanged();
   }
QPainterPath TShape::Outline() const
   {
      QPainterPath Path;
      Path.addRect(QRectF(BoundingRect().topLeft(), QSizeF(qMax(BoundingRect().width(), 1e-5), qMax(BoundingRect().height(), 1e-5))));
      return Path;
   }
void TShape::Rotate(qreal angle)
   {
      qreal Angle = Rotation();
      if (Angle == angle) return;
      Q_D(TShape);

      QPointF Pos = Position();
      d->Transformation.translate(Pos.x(), Pos.y());
      d->Transformation.rotate(angle - d->Rotation);
      d->Rotation = angle;
      d->Transformation.translate(-Pos.x(), -Pos.y());

      emit ShapeChanged();
   }
qreal TShape::Rotation() const
   {
      Q_D(const TShape);
      return d->Rotation;
   }
void TShape::ApplyAbsoluteTransformation(const QTransform &t)
   {
      QTransform GlobalTransform = AbsoluteTransformation();
      // the transformation is relative to the global coordinate system
      // but we want to change the local matrix, so convert the matrix
      // to be relative to the local coordinate system
      QTransform Transform = GlobalTransform * t * GlobalTransform.inverted();
      ApplyTransformation(Transform);
   }
QTransform TShape::AbsoluteTransformation() const
   {
      Q_D(const TShape);
      return d->Transformation;
   }
QSizeF TShape::size() const
   {
      //Q_D(const TShape);
      //return d->Size;
      //return d->Size == QSizeF() ? BoundingRect().size() : d->Size;
      return BoundingRect().size();
   }
void TShape::SetSize(const QSizeF &NewSize)
   {
      Q_D(TShape);
      QSizeF OldSize(size());
      //if (OldSize == NewSize) return;
      QTransform Matrix;
      OldSize.setHeight(qMax((qreal)1e-4, OldSize.height())); // avoids devision by zero below
      OldSize.setWidth(qMax((qreal)1e-4, OldSize.width()));
      d->Transformation.scale(NewSize.width() / OldSize.width(), NewSize.height() / OldSize.height());
      d->Size = NewSize;
      //ApplyAbsoluteTransformation(Matrix);
      emit ShapeChanged();
   }
void TShape::ApplyTransformation(const QTransform &t)
   {
      Q_D(TShape);
      d->Transformation = t * d->Transformation;
      emit ShapeChanged();
   }*/

}
