#include "TDefaultTool.h"
#include "tools/ToolManager.h"
#include "tools/DefaultTool/TMoveStrategy.h"
#include "tools/DefaultTool/TRotateStrategy.h"
#include "tools/DefaultTool/TResizeStrategy.h"
#include "tools/DefaultTool/TSelectDecorator.h"
#include "widgets/canvas/TWView.h"
#include "widgets/canvas/TConversion.h"

#include "shape/ShapeManager.h"
#include "shape/TSelectionShape.h"
#include "shape/TShape.h"
#include <math.h>

namespace graphed {

TDefaultTool::TDefaultTool(TCanvas *c): TInteractionTool(c)
{
    Icon = (makeHugeIcon(":/Lithography/hand_off.png"));
    ToolTip = (tr("Selection & move & resize tool"));
    setObjectName(ToolName());
}
GraphEd::THandles TDefaultTool::HandleAt(const TShapeSP Shape, const QPointF &DocumentPos, const bool DoSort, const unsigned int Dist)
{
    TConversionBase *Conversion = INSTANCE(TWView)->Conversion();
    if (!Conversion) return GraphEd::NoHandle;
    const QPointF WidgetPos = Conversion->DocumentToWidget(DocumentPos);

    //We should sort corners (in case they are flipped/mirrored)
    TPointFList Corners; // Corners of the shape in global coords;
    Corners << Shape->GlobalPosition(Shape->BoundingRect().topLeft());
    Corners << Shape->GlobalPosition(Shape->BoundingRect().topRight());
    Corners << Shape->GlobalPosition(Shape->BoundingRect().bottomRight());
    Corners << Shape->GlobalPosition(Shape->BoundingRect().bottomLeft());
    TPointFList HandlePointsinDocCoord = GraphEd::GetHandlePoints(Corners, DoSort);

    QRectF HandleRect(0, 0, 2.*Dist, 2.*Dist);
    TPointFList::size_type i = 0;
    foreach (QPointF Point, HandlePointsinDocCoord)
    {
        const QPointF WidgetPoint = Conversion->DocumentToWidget(Point);
        HandleRect.moveCenter(WidgetPoint);
        if (HandleRect.contains(WidgetPos)) return (GraphEd::THandles)i;
        ++i;
    }
    return (GraphEd::THandles)i;
}

TDefaultTool::TAction TDefaultTool::GetAction(const TShapeSP Shape, const QPointF &DocPos) const
{
    if (!Shape) return Act_None;

    GraphEd::THandles Handle = HandleAt(Shape, DocPos);

    //... and click was in it
    QPointF PosInShapesCoord = Shape->Transformation().inverted().map(DocPos);
    if (Shape->BoundingRect().contains(PosInShapesCoord))
    {
        if (Handle == GraphEd::NoHandle) return Act_Move;
        else return Act_Resize;
    }
    //Not implemented yet
    //if (Handle == GraphEd::TopLeftHandle || Handle == GraphEd::TopRightHandle || Handle == GraphEd::BottomRightHandle || Handle == GraphEd::BottomLeftHandle) return Act_Rotate;
    return Act_None;
}

void SetSelectionParams(qreal cx, qreal cy, qreal w, qreal h)
{
    if (!INSTANCE(ShapeManager)->Selection()->SelectedShapes().size())
        return;

    TShapeSP Shape = INSTANCE(ShapeManager)->Selection()->SelectedShapes().size() == 1 ? INSTANCE(ShapeManager)->Selection()->SelectedShapes().front() : (TShapeSP)INSTANCE(ShapeManager)->Selection();
    TConversionBase *Conversion = INSTANCE(TWView)->Conversion();

    QPointF pos = Shape->Position();
    QPointF TopLeft = Shape->GlobalPosition(Shape->BoundingRect().topLeft());
    QPointF BottomRight = Shape->GlobalPosition(Shape->BoundingRect().bottomRight());

    pos = Conversion->ResolutionToPhysical(Conversion->VirginVRToResolutionF(pos));
    TopLeft = Conversion->ResolutionToPhysical(Conversion->VirginVRToResolutionF(TopLeft));
    BottomRight = Conversion->ResolutionToPhysical(Conversion->VirginVRToResolutionF(BottomRight));

    QRectF rect(TopLeft, BottomRight);
    qreal y0 = Conversion->DocumentRect().top();
    qreal x0 = Conversion->DocumentRect().left();

    qreal deltax = x0 + cx*1e-6 - pos.x();
    qreal deltay = y0 + cy*1e-6 - pos.y();
    QPointF Delta = Conversion->PhysicalTo(Conversion->VirginViewRect(), QPointF(deltax, deltay));

    foreach (TShapeSP Shape0, INSTANCE(ShapeManager)->Selection()->SelectedShapes())
        Shape0->SetPosition(Shape0->Position() + Delta);

    qreal sx = w/(rect.width()*1e6);
    qreal sy = h/(rect.height()*1e6);

    foreach (TShapeSP Shape0, INSTANCE(ShapeManager)->Selection()->SelectedShapes())
    {
        QTransform t = Shape0->Transformation();

        const qreal XTrans = Shape->GlobalPosition().x();
        const qreal YTrans = Shape->GlobalPosition().y();

        t.translate(XTrans, YTrans);
        t.scale(sx, sy);
        t.translate(-XTrans, -YTrans);
        Shape0->SetTransformation(t);
    }

    INSTANCE(ShapeManager)->Selection()->InvalidateCache();
    INSTANCE(TCanvas)->update();
}

void UpdateSelectionParams()
{
    TConversionBase *Conversion = INSTANCE(TWView)->Conversion();
    TShapeSP Shape = INSTANCE(ShapeManager)->Selection()->SelectedShapes().size() == 1 ? INSTANCE(ShapeManager)->Selection()->SelectedShapes().front() : (TShapeSP)INSTANCE(ShapeManager)->Selection();

    QPointF pos = Shape->Position();
    QPointF TopLeft = Shape->GlobalPosition(Shape->BoundingRect().topLeft());
    QPointF BottomRight = Shape->GlobalPosition(Shape->BoundingRect().bottomRight());

    pos = Conversion->ResolutionToPhysical(Conversion->VirginVRToResolutionF(pos));
    TopLeft = Conversion->ResolutionToPhysical(Conversion->VirginVRToResolutionF(TopLeft));
    BottomRight = Conversion->ResolutionToPhysical(Conversion->VirginVRToResolutionF(BottomRight));

    QRectF rect(TopLeft, BottomRight);

    INSTANCE(ToolManager)->SetCurrParams(floor(pos.x()*1e6*1e3)/1e3,
                                         floor(pos.y()*1e6*1e3)/1e3,
                                         fabs(floor(rect.width()*1e6*1e3)/1e3),
                                         fabs(floor(rect.height()*1e6*1e3)/1e3));
}

TInteractionStrategy *TDefaultTool::CreateStrategy(TMouseEvent *e) const
{
    const bool CtrlPressed = e->modifiers() & Qt::ControlModifier;

    //If we already have some selection...
    if (!CtrlPressed && INSTANCE(ShapeManager)->Selection()->count())
    {
        TShapeSP Shape = INSTANCE(ShapeManager)->Selection()->SelectedShapes().size() == 1 ? INSTANCE(ShapeManager)->Selection()->SelectedShapes().front() : (TShapeSP)INSTANCE(ShapeManager)->Selection();
        TAction Action = GetAction(Shape, e->posF());

        UpdateSelectionParams();

        if (Action == Act_Move)        return new TMoveStrategy(Canvas, e);
        else if (Action == Act_Resize) return new TResizeStrategy(Canvas, e);
        else if (Action == Act_Rotate) return new TRotateStrategy(Canvas, e);
    }

    //select/deselect shapes...
    if (CtrlPressed)
    {
        TShapeSP Shape = INSTANCE(ShapeManager)->ShapeAt(e->posF(), GraphEd::ShapeOnTop);
        if (!Shape)
            return new TMoveStrategy(Canvas, e);

        if (!INSTANCE(ShapeManager)->Selection()->IsSelected(Shape))
            INSTANCE(ShapeManager)->Selection()->Select(Shape);
        else
            INSTANCE(ShapeManager)->Selection()->DeSelect(Shape);

        UpdateSelectionParams();

        Canvas->update();
        return new TMoveStrategy(Canvas, e);
    }
    else if (TShapeSP Shape = INSTANCE(ShapeManager)->ShapeAt(e->posF(), GraphEd::ShapeOnTop))
    {
        INSTANCE(ShapeManager)->Selection()->DeSelectAll();
        INSTANCE(ShapeManager)->Selection()->Select(Shape);

        UpdateSelectionParams();

        Canvas->update();
        return new TMoveStrategy(Canvas, e);
    }
    else
    {
        INSTANCE(ShapeManager)->Selection()->DeSelectAll();
        Canvas->update();
    }
    return NULL;
}
void TDefaultTool::paint(QPainter *p, const TConversionBase *Convert)
{
    //TInteractionTool::paint(p);
    if (INSTANCE(ShapeManager)->Selection()->count())
    {
        TSelectDecorator Decorator;
        Decorator.paint(p, Convert);
    }
}
void TDefaultTool::mouseMoveEvent(TMouseEvent *e)
{
    if (CurrentStrategy) {
        CurrentStrategy->handleMouseMove(e);
        UpdateSelectionParams();
    } else if (INSTANCE(ShapeManager)->Selection()->count()) {
        TShapeSP Shape = INSTANCE(ShapeManager)->Selection()->SelectedShapes().size() == 1 ? INSTANCE(ShapeManager)->Selection()->SelectedShapes().front() : (TShapeSP)INSTANCE(ShapeManager)->Selection();
        TAction Action = GetAction(Shape, e->posF());

        if (Action == Act_Move)        Canvas->setCursor(TMoveStrategy::Cursor());
        else if (Action == Act_Rotate) Canvas->setCursor(TRotateStrategy::Cursor());
        else if (Action == Act_Resize) Canvas->setCursor(TResizeStrategy::Cursor(e->posF()));
        else Canvas->unsetCursor();
    } else {
        Canvas->unsetCursor();
        INSTANCE(ToolManager)->SetCurrParams(0, 0, 1, 1);
    }
}
void TDefaultTool::Deactivate()
{
    if (INSTANCE(ShapeManager)->Selection()->count())
    {
        INSTANCE(ShapeManager)->Selection()->DeSelectAll();
        Canvas->update();
    }
}
void TDefaultTool::on_DeletePressed()
{
    delete CurrentStrategy; CurrentStrategy = NULL;
    if (INSTANCE(ShapeManager)->Selection()->count())
    {
        INSTANCE(ShapeManager)->Remove(INSTANCE(ShapeManager)->Selection()->SelectedShapes());
        INSTANCE(ShapeManager)->Selection()->DeSelectAll();
        Canvas->unsetCursor();
        Canvas->update();
    }
}
void TDefaultTool::keyPressEvent(QKeyEvent *e)
{
    if (e->key() == Qt::Key_Delete) on_DeletePressed();
}

}

