#include "TResizeStrategy.h"
#include "tools/DefaultTool/TDefaultTool.h"
#include <limits>
#include <math.h>

namespace graphed {

TResizeStrategy::TResizeStrategy(const TCanvasQP c, TMouseEvent *e): TInteractionStrategy(c, e->posF())
{
    if (!INSTANCE(ShapeManager)->Selection()->count()) return;
    Canvas->setCursor(Cursor(e->posF()));

    TShapeSP Shape = SelectedShapes.size() == 1 ? SelectedShapes.front() : (TShapeSP)INSTANCE(ShapeManager)->Selection();
    if (SelectedShapes.size() > 1) SelectedShapes << INSTANCE(ShapeManager)->Selection();

    //Define Points
    Center = Shape->BoundingRect().center();
    ActiveHandle = TDefaultTool::HandleAt(Shape, e->posF(), false);

    TPointFList HandlePoints = GraphEd::GetHandlePoints(Shape->BoundingRect());
    switch(ActiveHandle)
    {
    case GraphEd::TopLeftHandle:
        FixedCorner = HandlePoints[GraphEd::BottomRightHandle];
        Top = true; Bottom = false; Left = true; Right = false; break;
    case GraphEd::TopMiddleHandle:
        FixedCorner = HandlePoints[GraphEd::BottomMiddleHandle];
        Top = true; Bottom = false; Left = false; Right = false; break;
    case GraphEd::TopRightHandle:
        FixedCorner = HandlePoints[GraphEd::BottomLeftHandle];
        Top = true; Bottom = false; Left = false; Right = true; break;
    case GraphEd::RightMiddleHandle:
        FixedCorner = HandlePoints[GraphEd::LeftMiddleHandle];
        Top = false; Bottom = false; Left = false; Right = true; break;
    case GraphEd::BottomRightHandle:
        FixedCorner = HandlePoints[GraphEd::TopLeftHandle];
        Top = false; Bottom = true; Left = false; Right = true; break;
    case GraphEd::BottomMiddleHandle:
        FixedCorner = HandlePoints[GraphEd::TopMiddleHandle];
        Top = false; Bottom = true; Left = false; Right = false; break;
    case GraphEd::BottomLeftHandle:
        FixedCorner = HandlePoints[GraphEd::TopRightHandle];
        Top = false; Bottom = true; Left = true; Right = false; break;
    case GraphEd::LeftMiddleHandle:
        FixedCorner = HandlePoints[GraphEd::RightMiddleHandle];
        Top = false; Bottom = false; Left = true; Right = false; break;
    }
    FixedPoint = (StartSizeWasCalcedRelativeCenter = e->modifiers() & Qt::ControlModifier) ? Center : FixedCorner;

    //Define OldTransforms
    foreach (TShapeSP Shape, SelectedShapes) OldTransforms << Shape->Transformation();
    //and StartSize
    StartSize = QSizeF(Shape->GlobalPosition(FixedPoint).x() - e->posF().x(), Shape->GlobalPosition(FixedPoint).y() - e->posF().y());
}
TResizeStrategy::~TResizeStrategy()
{
    Canvas->unsetCursor();
}
void TResizeStrategy::UpdateStartSize(const bool ControlModifier)
{
    ControlModifier ? StartSize /= 2. : StartSize *= 2.;
    FixedPoint = ControlModifier ? Center : FixedCorner;
    StartSizeWasCalcedRelativeCenter = ControlModifier;
}

void TResizeStrategy::CancelInteraction()
{
    TShapeList::size_type i = 0;
    foreach (TShapeSP Shape, SelectedShapes) Shape->SetTransformation(OldTransforms[i++]);
    Canvas->update();
}
void TResizeStrategy::handleMouseMove(TMouseEvent *e)
{
    if (!SelectedShapes.size()) return;

    if (StartSizeWasCalcedRelativeCenter != (bool)(e->modifiers() & Qt::ControlModifier))
        UpdateStartSize(e->modifiers() & Qt::ControlModifier);

    //Define sx and sy
    QPointF CurrentPos = e->posF();

    TShapeSP Shape = SelectedShapes.size() == 1 ? SelectedShapes.front() : (TShapeSP)INSTANCE(ShapeManager)->Selection();

    qreal NewWidth = Shape->GlobalPosition(FixedPoint).x() - CurrentPos.x();
    qreal NewHeight = Shape->GlobalPosition(FixedPoint).y() - CurrentPos.y();

    qreal sx = NewWidth / StartSize.width();
    qreal sy = NewHeight / StartSize.height();

    sx = qMax(std::numeric_limits<qreal>::epsilon(), fabs(sx)) * GraphEd::Sign(sx);
    sy = qMax(std::numeric_limits<qreal>::epsilon(), fabs(sy)) * GraphEd::Sign(sy);

    if (ActiveHandle == GraphEd::RightMiddleHandle    || ActiveHandle == GraphEd::LeftMiddleHandle)
        sy = 1.;
    else if (ActiveHandle == GraphEd::TopMiddleHandle || ActiveHandle == GraphEd::BottomMiddleHandle)
        sx = 1.;

    const bool KeepAspect = e->modifiers() & Qt::ShiftModifier;
    if (KeepAspect)
    {
        const bool CornerUsed = ((Bottom ? 1 : 0) + ( Top ? 1 : 0) + (Left ? 1 : 0) + (Right ? 1 : 0)) == 2;
        if ((CornerUsed && NewWidth < NewHeight) || Left || Right) sy = sx;
        else sx = sy;
    }

    //Apply sx and sy to Shapes
    QPointF FixPointInGlobalCoord = OldTransforms.back().map(FixedPoint);
    TShapeList::size_type i = 0;
    foreach (TShapeSP Shape, SelectedShapes)
    {
        QTransform t = OldTransforms[i++];

        QPointF FixPointInShapesCoord = t.inverted().map(FixPointInGlobalCoord);
        const qreal XTrans = FixPointInShapesCoord.x();
        const qreal YTrans = FixPointInShapesCoord.y();

        t.translate(XTrans, YTrans);
        t.scale(sx, sy);
        t.translate(-XTrans, -YTrans);
        Shape->SetTransformation(t);
    }

    Canvas->update();
}
QCursor TResizeStrategy::Cursor(const QPointF &Pos)
{
    if (!INSTANCE(ShapeManager)->Selection()->count()) return QCursor();
    TShapeSP Shape = INSTANCE(ShapeManager)->Selection()->count() == 1 ? INSTANCE(ShapeManager)->Selection()->SelectedShapes().front() : (TShapeSP)INSTANCE(ShapeManager)->Selection();

    GraphEd::THandles Handle = TDefaultTool::HandleAt(Shape, Pos);

    if (Handle == GraphEd::TopLeftHandle     || Handle == GraphEd::BottomRightHandle)  return QCursor(Qt::SizeFDiagCursor);
    if (Handle == GraphEd::TopRightHandle    || Handle == GraphEd::BottomLeftHandle)   return QCursor(Qt::SizeBDiagCursor);
    if (Handle == GraphEd::TopMiddleHandle   || Handle == GraphEd::BottomMiddleHandle) return QCursor(Qt::SizeVerCursor);
    if (Handle == GraphEd::RightMiddleHandle || Handle == GraphEd::LeftMiddleHandle)   return QCursor(Qt::SizeHorCursor);
    else return QCursor();
}
}
