#include "TApplyTool.h"
#include "TApplyTool_p.h"
#include "widgets/canvas/TConversionBase.h"
#include "out/TConvertor.h"
#include "widgets/canvas/TWCanvasController.h"
#include "core/core.h"
#include "shape/ShapeManager.h"
#include "tools/ToolManager.h"

namespace graphed {

int TApplyToolThreadInfo::Progress() const
{
    QMutexLocker lck(&AccessControl);
    return m_Progress;
}
void TApplyToolThreadInfo::SetProgress(const int P)
{
    AccessControl.lock();
    m_Progress = qBound(0, P, 100);
    AccessControl.unlock();
}
QString TApplyToolThreadInfo::CurrentOperation() const
{
    QMutexLocker lck(&AccessControl);
    return m_CurrentOperation;
}
void TApplyToolThreadInfo::SetCurrentOperation(const QString &Text)
{
    AccessControl.lock();
    m_CurrentOperation = Text;
    AccessControl.unlock();
}
TShapeSP TApplyToolThreadInfo::CurrentShape() const
{
    QMutexLocker lck(&AccessControl);
    return m_CurrentShape;
}
void TApplyToolThreadInfo::SetCurrentShape(const TShapeSP Shape)
{
    AccessControl.lock();
    m_CurrentShape = Shape;
    AccessControl.unlock();
}
void TApplyToolThreadInfo::ResetAllProgressInfo()
{
    AccessControl.lock();
    m_CurrentShape.clear();
    m_CurrentOperation.clear();
    m_Progress = 0;
    AccessControl.unlock();
}

//=======================================================================================================================

TApplyToolThread::TApplyToolThread(TApplyToolPrivate *qq): parent(qq)
{
}
void TApplyToolThread::MakeOutputResults()
{
    SetCurrentOperation("MakeOutputResults");
    parent->ResultsList.clear();
    TShapeList Sorted = INSTANCE(ShapeManager)->ShapeList();
    qSort(Sorted.begin(), Sorted.end(), TShape::CompareShapeByZIndex);

    foreach (TShapeSP CurrSh, Sorted)
    {
        if (CurrSh->objectName() == GraphEd::BackgroundShapeName()) continue;
        SetCurrentOperation("Image proccessing..." ); SetCurrentShape(CurrSh);
        if (out::TResults *Res = CurrSh->ToPointList())
        {
            if (Res->PointList().size() != 0)
                parent->ResultsList << Res;
        }
    }
    ResetAllProgressInfo();
}
void TApplyToolThread::MakeImage()
{
    SetCurrentOperation("MakeImage");
    parent->Shape.clear();
    parent->Shape = TImageShapeSP(new TImageShape);
    parent->Shape->setObjectName(GraphEd::ResultShapeName());
    parent->Shape->SetSelectable(false);

    QImage *Image = dynamic_cast<QImage *>(parent->Shape.data());
    if (!Image) return;
    const TConversionBase *Conversion = parent->parent->Canvas->Conversion();
    if (!Conversion) return;

    const unsigned int TotalSteps = parent->ResultsList.size();
    unsigned int CurrentStep = 0;

    const QRectF ToIt = QRectF(QPointF(), Conversion->DocumentResolution());
    *Image = QImage(Conversion->DocumentResolution(), QImage::Format_ARGB32_Premultiplied);
    Image->fill(0);
    {
        QPainter p(Image);
        for (out::TResultsList::const_iterator ResIt = parent->ResultsList.begin(); ResIt != parent->ResultsList.end(); ResIt++)
        {
            SetProgress((CurrentStep++ * 100) / TotalSteps);
            const STLPointFVector &List = (*ResIt)->PointList();
            for (STLPointFVector::const_iterator PIt = List.begin(); PIt != List.end(); PIt++)
            {
                QPen Pen = QPen(Qt::red);
                const out::TVarVoltage *VarVoltage = dynamic_cast<const out::TVarVoltage *>(*ResIt);
                if (VarVoltage) Pen = out::TVoltColourConvertor::Voltage2Colour((*ResIt)->VtData1(PIt).first, 0, VarVoltage->MaxV());
                p.setPen(Pen);
                const QPointF &Point = Conversion->PhysicalTo(ToIt, *PIt);
                p.drawPoint(Point.toPoint());
            }
        }
    }
    *Image = Image->scaled(Conversion->VirginViewRect().size().toSize(), Qt::IgnoreAspectRatio, Qt::SmoothTransformation);

    ResetAllProgressInfo();
}
void TApplyToolThread::run()
{
    MakeOutputResults();
    MakeImage();
}
//=======================================================================================================================

TApplyToolPrivate::TApplyToolPrivate()
{
    Timer.setInterval(100);
    Thread = new TApplyToolThread(this);
    connect(this, SIGNAL(SetProgress(int,QString)), INSTANCE(TCanvas), SLOT(SetProgress(int,QString)));
    connect(&Timer, SIGNAL(timeout()), this, SLOT(on_Timer_timeout()));
    connect(Thread, SIGNAL(started()), this, SLOT(on_started()));
    connect(Thread, SIGNAL(finished()), this, SLOT(on_finished()));

}
QString TApplyToolPrivate::ParamsToString(const int X, const int Y) const
{
    return QString("(X = ") + QString::number(X) + QString("; Y = ") + QString::number(Y) + ")";
}

void TApplyToolPrivate::on_started()
{
    emit parent->SetLockRequest(true);
    parent->Canvas->UpdateCanvas();
    Timer.start();
}
void TApplyToolPrivate::on_finished()
{
    Timer.stop();
    emit parent->SetLockRequest(false);

    FinishShape(Shape);
    parent->Canvas->UpdateCanvas();

    INSTANCE(ToolManager)->GetButtonsOfTools().at(0)->click(); //activate interacttool
}
void TApplyToolPrivate::on_Timer_timeout()
{
    TShapeSP Shape = Thread->CurrentShape();
    const int Progress = Shape ? Shape->Progress() : Thread->Progress();
    emit SetProgress(Progress, Thread->CurrentOperation());
}

//=======================================================================================================================

TApplyTool::TApplyTool(TCanvas *c): TTool(c)
{
    child = new TApplyToolPrivate;
    child->parent = this;
    Icon = (makeHugeIcon(":/Lithography/Apply.png"));
    ToolTip = (QObject::tr("Convert to points"));
    setObjectName(ToolName());
}
TApplyTool::~TApplyTool()
{
    delete child;
}
void TApplyTool::Activate()
{
    INSTANCE(ShapeManager)->RemoveAll(GraphEd::ResultShapeName());
    child->Thread->start();
}
const out::TResultsList &TApplyTool::ResultsList() const
{
    return child->ResultsList;
}
void TApplyTool::ClearResults()
{
    child->ResultsList.clear();
}
}


