#include "ToolManager.h"
#include "ToolManager_p.h"
#include "tools/TPointTool.h"
#include "tools/TLineTool.h"
#include "tools/TRectTool.h"
#include "tools/TImageTool.h"
#include "tools/TApplyTool.h"
#include "tools/TEllipseTool.h"
#include "tools/TEllipseCurveTool.h"
#include "tools/DefaultTool/TDefaultTool.h"
#include "tools/TPosTool.h"

#include <QDebug>
#include <stddef.h>

graphed::TTool *FindTool(const QString name)
{
    return dynamic_cast<graphed::TTool *>(INSTANCE(graphed::ToolManager)->AllTools().find(name).value());
}

namespace graphed {

ToolManagerPrivate::ToolManagerPrivate()
   {
      Canvas = INSTANCE(TCanvas);
      ToolProxy = Canvas->ToolProxy();

      AddTool(new TDefaultTool(Canvas));

      AddTool(new TLineTool(Canvas));
      AddTool(new TRectTool(Canvas));
      AddTool(new TEllipseTool(Canvas));
      AddTool(new TEllipseCurveTool(Canvas));
      AddTool(new TImageTool(Canvas));
      AddTool(new TPointTool(Canvas));

      AddTool(new TPosTool(Canvas));
      AddTool(new TApplyTool(Canvas));

      for (ToolManager::TToolMap::const_iterator It = AllTools.begin(); It != AllTools.end(); It++)
         connect(It.value(), SIGNAL(ToolActivated(const QString &)), this, SLOT(on_ToolActivated(const QString &)));
   }
ToolManagerPrivate::~ToolManagerPrivate()
   {
      for (ToolManager::TToolMap::const_iterator It = AllTools.begin(); It != AllTools.end(); It++)
         delete It.value();
      AllTools.clear();
   }

void ToolManagerPrivate::AddTool(TTool *Tool)
   {
      if (AllTools.find(Tool->objectName()) != AllTools.end())
         { qDebug() << "void ToolManagerPrivate::AddTool(TTool *Tool): Tool " << Tool->objectName() << " already exists!"; return; }
      AllTools[Tool->objectName()] = Tool;
   }
void ToolManagerPrivate::on_ToolActivated(const QString &Name)
   {
      ToolManager::TToolMap::const_iterator It = AllTools.find(Name);
      if (It != AllTools.end())
         {
            ToolProxy->Deactivate();
            ToolProxy->SetActiveTool(It.value());
            ToolProxy->Activate();
         }
      else qDebug() << "Can not find tool" << Name;
   }

void ToolManagerPrivate::AddToolToButtonsOfTools(const QString &Name)
   {
      ToolManager::TToolMap::const_iterator It = AllTools.find(Name);
      if (It == AllTools.end()) return;

      QPushButton *B = new QPushButton(It.value()->Icon, QString());
      B->setToolTip(It.value()->ToolTip);
      B->setCheckable(true);
      QObject::connect(B, SIGNAL(clicked(bool)), It.value(), SLOT(on_ToolButton_clicked(bool)));
      ButtonsOfTools.push_back(B);
   }
const ToolManager::TBList &ToolManagerPrivate::CreateButtonsOfTools()
   {
      AddToolToButtonsOfTools(TDefaultTool::ToolName());

      AddToolToButtonsOfTools(TPointTool::ToolName());
      AddToolToButtonsOfTools(TLineTool::ToolName());
      AddToolToButtonsOfTools(TRectTool::ToolName());
      AddToolToButtonsOfTools(TEllipseTool::ToolName());
      AddToolToButtonsOfTools(TEllipseCurveTool::ToolName());
      AddToolToButtonsOfTools(TImageTool::ToolName());

      AddToolToButtonsOfTools(TPosTool::ToolName());
      AddToolToButtonsOfTools(TApplyTool::ToolName());
      return ButtonsOfTools;
   }
void ToolManagerPrivate::ConnectSlots()
   {
      Q_Q(ToolManager);
      for (ToolManager::TToolMap::const_iterator It = AllTools.begin(); It != AllTools.end(); It++)
         connect(It.value(), SIGNAL(SetLockRequest(bool)), q, SIGNAL(SetLockRequest(bool)));
   }

//=====================================================================================================================

ToolManager *ToolManager::Self = NULL;
bool ToolManager::IAmBeingCreatedNow = false;
ToolManager::ToolManager(): d_ptr(new ToolManagerPrivate)
   {
      Q_D(ToolManager);
      d->q_ptr = this;
      d->ConnectSlots();
   }
ToolManager::~ToolManager() { delete d_ptr; }
ToolManager *ToolManager::Instance(const int Line, const char *File)
   {
      if (IAmBeingCreatedNow)
         { qDebug() << "ToolManager::Instance() was called from " << File << " and line #" << Line << "! Returning NULL..."; return NULL; }
      if (!Self)
         {
            IAmBeingCreatedNow = true;
            Self = new ToolManager;
            IAmBeingCreatedNow = false;
         }
      return Self;
   }
void ToolManager::Destroy()  { delete Self; Self = 0;  }
//=====================================================================================================================
const ToolManager::TBList &ToolManager::GetButtonsOfTools()
   {
      Q_D(ToolManager);
      if (d->ButtonsOfTools.size()) return d->ButtonsOfTools;
      return d->CreateButtonsOfTools();
   }
TToolProxy *ToolManager::GetToolProxy() const
   {
      Q_D(const ToolManager);
      return d->ToolProxy;
   }
const ToolManager::TToolMap &ToolManager::AllTools() const
   {
      Q_D(const ToolManager);
      return d->AllTools;
   }

}
