#include "TWToolBox.h"
#include "tools/ToolManager.h"
#include "tools/TApplyTool.h"
#include "core/core.h"
#include "shape/ShapeManager.h"
#include "widgets/canvas/TWView.h"
#include "widgets/canvas/TConversion.h"
#include "flowlayout.hpp"
#include <QShortcut>
#include <QDebug>

namespace graphed {

TWToolBox::TWToolBox(QWidget *parent) : QWidget(parent)
{
    fullrect = INSTANCE(TWView)->Conversion()->DocumentRect();
    is_rect_set = false;

    QButtonGroup *ButtonGroup = new QButtonGroup(this);
    HFlowLayout *MainLayout = new HFlowLayout(this);
    MainLayout->spaceX = MainLayout->spaceY = 5;
    MainLayout->setContentsMargins(0, 0, 0, 0);
    MainLayout->setAlignment(Qt::AlignLeft);

    QPushButton *BSetScale = new QPushButton(makeHugeIcon(":/Lithography/selection.png"), QString());
    BSetScale->setToolTip(QObject::tr("Save current ranges and background"));
    QPushButton *BRestoreScale = new QPushButton(makeHugeIcon(":/Lithography/restore_scale.png"), QString());
    BRestoreScale->setToolTip(QObject::tr("Restore saved ranges and background"));
    QPushButton *BClearAll = new QPushButton(makeHugeIcon(":/Lithography/clear.png"), QString());
    BClearAll->setToolTip(tr("Clear All"));

    const ToolManager::TBList &List = INSTANCE(ToolManager)->GetButtonsOfTools();
    for (ToolManager::TBList::const_iterator It = List.begin(); It != List.end(); It++)
    {
        MainLayout->addWidget(*(It));
        ButtonGroup->addButton(*(It));
        blist << *(It);
    }
    addDelimeterH(MainLayout, BSetScale->sizeHint().width());
    MainLayout->addWidget(BSetScale);
    MainLayout->addWidget(BRestoreScale);
    MainLayout->addWidget(BClearAll);
    blist << BSetScale << BRestoreScale << BClearAll;

    new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_A), this, SLOT(selectAll()));
    QObject::connect(BSetScale, SIGNAL(clicked()), this, SLOT(BSetScale_clicked()));
    QObject::connect(BRestoreScale, SIGNAL(clicked()), this, SLOT(BRestoreScale_clicked()));
    QObject::connect(BClearAll, SIGNAL(clicked()), this, SLOT(on_BClear_clicked()));
}

void TWToolBox::SetLocked(const bool Lock)
{
    foreach (QWidget *B, blist)
        B->setEnabled(!Lock);
}
void TWToolBox::on_BClear_clicked()
{
    TApplyTool *ApplyTool = (TApplyTool *)FindTool("TApplyTool");
    ApplyTool->ClearResults();
    INSTANCE(ShapeManager)->RemoveAllExcept(GraphEd::BackgroundShapeName());
    if (TToolProxy *TP = INSTANCE(TCanvas)->ToolProxy()) TP->Deactivate();
    INSTANCE(TCanvas)->UpdateCanvas();
}
void TWToolBox::selectAll()
{
    for (int i = 0; i < INSTANCE(ShapeManager)->count(); i++)
        INSTANCE(ShapeManager)->Selection()->Select(INSTANCE(ShapeManager)->ShapeList().at(i));
    INSTANCE(TCanvas)->UpdateCanvas();
}

void TWToolBox::BSetScale_clicked()
{
    TConversionBase *Conversion = INSTANCE(TWView)->Conversion();

    if (!is_rect_set)
        fullrect = Conversion->DocumentRect();
    is_rect_set = true;

    //find new physical ranges
    QPointF TopLeft = Conversion->ViewRect().topLeft();
    TopLeft.rx() = qBound<qreal>(0., TopLeft.x(), static_cast<qreal>(INSTANCE(TCanvas)->width()));
    TopLeft.ry() = qBound<qreal>(0., TopLeft.y(), static_cast<qreal>(INSTANCE(TCanvas)->height()));

    QPointF BottomRight = Conversion->ViewRect().bottomRight();
    BottomRight.rx() = qBound<qreal>(0., BottomRight.x(), static_cast<qreal>(INSTANCE(TCanvas)->width()));
    BottomRight.ry() = qBound<qreal>(0., BottomRight.y(), static_cast<qreal>(INSTANCE(TCanvas)->height()));

    TopLeft = Conversion->ViewToPhysical(TopLeft);
    BottomRight = Conversion->ViewToPhysical(BottomRight);

    //find rect for new background
    QPointF TopLeft2 = Conversion->PhysicalTo(Conversion->VirginViewRect(), TopLeft);
    QPointF BottomRight2 = Conversion->PhysicalTo(Conversion->VirginViewRect(), BottomRight);

    QRectF rc(TopLeft2, BottomRight2);

    TShapeList List = INSTANCE(ShapeManager)->FindAll(GraphEd::BackgroundShapeName());
    if (!List.size()) {qDebug() << "Background not found!"; return;}
    QImage *ImgPtr = dynamic_cast<QImage *>(List.front().data());
    if (!ImgPtr) {qDebug() << "Background not found!"; return;}

    QImage newbg = ImgPtr->copy(rc.toRect());

    Conversion->SetOffset(QPointF());
    Conversion->SetZoom(1);

    INSTANCE(TWView)->CanvasController()->SetDocumentRect(QRectF(TopLeft, BottomRight).normalized());

    *ImgPtr = newbg.scaled(Conversion->VirginViewRect().size().toSize(), Qt::IgnoreAspectRatio, Qt::SmoothTransformation);

    INSTANCE(TCanvas)->UpdateCanvas();
}
void TWToolBox::BRestoreScale_clicked()
{
    if (is_rect_set) {
        is_rect_set = false;
        INSTANCE(TWView)->CanvasController()->SetDocumentRect(fullrect);
    }

    qreal zoom = INSTANCE(TWView)->Conversion()->Zoom();
    QRectF rect = INSTANCE(TWView)->Conversion()->ViewRect();

    INSTANCE(TWView)->Conversion()->SetOffset(
                INSTANCE(TWView)->Conversion()->Offset() -
                rect.topLeft());

    INSTANCE(TWView)->Conversion()->ZoomBy(QPointF(), 1/zoom);
    INSTANCE(TWView)->update_ruler();
    INSTANCE(TCanvas)->UpdateCanvas();
}

}
