#include "TCanvas.h"

#include <QDebug>
#include <QPainter>
#include "shape/ShapeManager.h"
#include "tools/ToolManager.h"
#include <QTime>
#include "widgets/canvas/TConversionBase.h"
#include "core/TMouseEvent.h"
#include "tools/TToolProxy.h"
#include "core/core.h"

namespace graphed {

class TCanvasPrivate
   {
      Q_DECLARE_PUBLIC(TCanvas);
   public:
      TCanvas * q_ptr;
      TCanvasPrivate();
      virtual ~TCanvasPrivate();
      void PaintProgress(QPainter *p);

      bool Locked;
      int Progress;
      QString ProgressText;

      QColor BackgroundColor;
      TToolProxy *ToolProxy;
      TConversionBase *ConversionBase;
      QPointF MarginOffset;
   };

TCanvasPrivate::TCanvasPrivate(): Locked(false), Progress(0), ConversionBase(NULL)
   {
      BackgroundColor = Qt::white;
      ToolProxy = new TToolProxy;
   }
TCanvasPrivate::~TCanvasPrivate() { }

void TCanvasPrivate::PaintProgress(QPainter *p)
   {
      Q_Q(TCanvas);
      QFont Font = p->font(); Font.setBold(true); Font.setPointSizeF(20);
      QFontMetricsF FontMetrics(Font);
      p->save();
      p->setCompositionMode(QPainter::CompositionMode_Multiply);
      p->setRenderHint(QPainter::Antialiasing);
      p->setBrush(QColor::fromRgba(qRgba(128, 128, 128, 200)));
      p->drawRect(q->rect());
      p->setFont(Font);

      const qreal Margin = 10;
      const qreal PercentWidth = 5;
      const qreal PercentHeight = 25;
      const QString UpperText("Please wait...");
      const QString DownText(ProgressText);
      const QString PercentText = QString::number(Progress) + "%";
      QRectF UpperRect = FontMetrics.boundingRect(UpperText).adjusted(-Margin, -Margin, Margin, Margin);
      QRectF DownRect = FontMetrics.boundingRect(DownText).adjusted(-Margin, -Margin, Margin, Margin);;
      QRectF TextPercentRect = FontMetrics.boundingRect(PercentText).adjusted(-Margin, -Margin, Margin, Margin);;
      QRectF PercentRect = QRectF(0, 0, Progress * PercentWidth, PercentHeight).adjusted(-Margin, -Margin, Margin, Margin);
      const qreal TotalHeight = (UpperRect.height() + qMax(PercentRect.height(), TextPercentRect.height()) + DownRect.height()) / 2.;
      //const qreal TotalWidth = qMax(UpperRect.width(), qMax(TextPercentRect.width() + PercentRect.width(), DownRect.width())) / 2.;
      UpperRect.moveTopLeft(QPointF(-UpperRect.width() / 2., -TotalHeight));
      PercentRect.moveTopLeft(QPointF(-(PercentRect.width() + TextPercentRect.width()) / 2., -TotalHeight + UpperRect.height()));
      TextPercentRect.moveTopLeft(QPointF(PercentRect.right(), -TotalHeight + UpperRect.height() + (PercentRect.height() - TextPercentRect.height()) / 2.));
      DownRect.moveTopLeft(QPointF(-DownRect.width() / 2., TotalHeight - DownRect.height()));

      QRectF BoundRect = UpperRect;
      BoundRect |= PercentRect;
      BoundRect |= TextPercentRect;
      BoundRect |= DownRect;
      BoundRect.adjust(-Margin, -Margin, Margin, Margin);

      p->translate(q->rect().center());
      p->setCompositionMode(QPainter::CompositionMode_Multiply);

      p->setPen(Qt::NoPen);
      p->setBrush(QColor(170, 230, 255).darker(180));
      p->drawRoundedRect(BoundRect, 30, 30);

      const QColor TextColour(Qt::white);
      p->setCompositionMode(QPainter::CompositionMode_SourceOver);
      p->setPen(TextColour);
      p->drawText(UpperRect, UpperText);
      p->setPen(QColor(Qt::green).lighter());
      p->setBrush(Qt::green);
      for (int i = 0; i < Progress; i++)
         p->drawRect(i * PercentWidth + PercentRect.left(), PercentRect.top(), PercentWidth / 2.,PercentHeight);
      p->setPen(TextColour);
      p->drawText(TextPercentRect, PercentText);
      p->drawText(DownRect, DownText);

      p->restore();
   }

//======================================================================================================================

TCanvas *TCanvas::Self = NULL;
bool TCanvas::IAmBeingCreatedNow = false;
TCanvas::TCanvas(): d_ptr(new TCanvasPrivate)
   {
      Q_D(TCanvas);
      d->q_ptr = this;
      resize(minimumSize());
      setMouseTracking(true);
      setFocusPolicy(Qt::ClickFocus);
   }
TCanvas::~TCanvas() { delete d_ptr;  }
TCanvas *TCanvas::Instance(const int /*Line*/, const char * /*File*/)
   {
      if (IAmBeingCreatedNow)
         { return NULL; }
      if (!Self)
         {
            IAmBeingCreatedNow = true;
            Self = new TCanvas;
            IAmBeingCreatedNow = false;
         }
      return Self;
   }
void TCanvas::Destroy()  { delete Self; Self = NULL; }
void TCanvas::SetConversion(TConversionBase *Conv)
   {
      Q_D(TCanvas);
      d->ConversionBase = Conv;
      d->MarginOffset = QPointF(d->ConversionBase->Margin(), d->ConversionBase->Margin());
   }
TConversionBase *TCanvas::Conversion() const
   {
      Q_D(const TCanvas);
      return d->ConversionBase;
   }
TToolProxy *TCanvas::ToolProxy() const
   {
      Q_D(const TCanvas);
      return d->ToolProxy;
   }
void TCanvas::resizeEvent(QResizeEvent * /*e*/)
   {
      update();
   }
void TCanvas::UpdateCanvas()
   {
      //possible work to take into the account zoom of whole document
      //QRect clipRect( viewToWidget( d->zoomHandler.documentToView(rc).toRect() ) );
      //clipRect.adjust(-2, -2, 2, 2); // grow for anti-aliasing
      //clipRect.moveTopLeft( clipRect.topLeft() - d->documentOffset);
      update();
   }
void TCanvas::paintEvent(QPaintEvent * /*e*/)
   {
      Q_D(TCanvas);
      QPainter p(this);
      p.setRenderHint(QPainter::Antialiasing);
      p.fillRect(QRectF(0, 0, size().width(), size().height()), QBrush(d->BackgroundColor));

      if (d->ConversionBase)
         {
            p.save();
               const QPointF MarginOffset(d->ConversionBase->Margin(), d->ConversionBase->Margin());
               p.translate(d->MarginOffset);
               p.drawRect(d->ConversionBase->ViewRect());
               /* TShape::ApplyConversion(&p, d->ConversionBase);
                  p.drawRect(QRect(QPoint(0., 0.), d->ConversionBase->DocumentResolution()));*/
            p.restore();
         }
      p.save();
         p.translate(d->ConversionBase->Offset() + d->MarginOffset);
         INSTANCE(ShapeManager)->paint(&p, d->ConversionBase);
         INSTANCE(ToolManager)->GetToolProxy()->paint(&p, d->ConversionBase);
      p.restore();
      if (d->Locked) d->PaintProgress(&p);
   }
void TCanvas::mousePressEvent(QMouseEvent *e)
   {
      Q_D(TCanvas);
      const QPointF PosWithoutMargin = e->posF() - d->MarginOffset;
      TMouseEvent E(e, d->ConversionBase ? d->ConversionBase->WidgetToDocument(PosWithoutMargin) : PosWithoutMargin);
      d->ToolProxy->mousePressEvent(&E);
   }
void TCanvas::mouseMoveEvent(QMouseEvent *e)
   {
      Q_D(TCanvas);
      const QPointF PosWithoutMargin = e->posF() - d->MarginOffset;
      TMouseEvent E(e, d->ConversionBase ? d->ConversionBase->WidgetToDocument(PosWithoutMargin) : PosWithoutMargin);
      d->ToolProxy->mouseMoveEvent(&E);
   }
void TCanvas::mouseReleaseEvent(QMouseEvent *e)
   {
      Q_D(TCanvas);
      const QPointF PosWithoutMargin = e->posF() - d->MarginOffset;
      TMouseEvent E(e, d->ConversionBase ? d->ConversionBase->WidgetToDocument(PosWithoutMargin) : PosWithoutMargin);
      d->ToolProxy->mouseReleaseEvent(&E);
   }
void TCanvas::keyReleaseEvent(QKeyEvent *e)
   {
      Q_D(TCanvas);
      d->ToolProxy->keyReleaseEvent(e);
   }
void TCanvas::keyPressEvent(QKeyEvent *e)
   {
      Q_D(TCanvas);
      d->ToolProxy->keyPressEvent(e);
   }
void TCanvas::SetLocked(const bool Lock)
   {
      Q_D(TCanvas);
      d->Locked = Lock;
      d->Progress = 0;
      UpdateCanvas();
   }
void TCanvas::SetProgress(const int Progress, const QString &Text)
   {
      Q_D(TCanvas);
      d->Progress = qBound(0, Progress, 100);
      d->ProgressText = Text;
      UpdateCanvas();
   }
/*QPointF TCanvas::MarginOffset() const
   {
      Q_D(const TCanvas);
      return d->MarginOffset;
   }*/

}
