#include "TWCanvasController.h"
#include "TWCanvasController_p.h"
#include "widgets/canvas/TWView.h"
#include "widgets/canvas/TCanvas.h"
#include "widgets/canvas/TConversion.h"
#include "core/core.h"
#include "shape/ShapeManager.h"
#include "shape/TImageShape.h"
#include "misc/TMisc.h"
#include <math.h>
#include <QScrollBar>
#include <QDebug>
#include <QWheelEvent>

namespace graphed {

TViewportWidget::TViewportWidget(TCanvas *c, TConversion *conv): Canvas(c), Conversion(conv)
   {
      Canvas->setParent(this);
      ResetLayout();
   }
TViewportWidget::~TViewportWidget() { }
void TViewportWidget::ResetLayout()
   {
      QRect ViewportRect(0, 0, width(), height());
      Canvas->setGeometry(ViewportRect);
   }
void TViewportWidget::ZoomBy(const QPointF &Center, const qreal Zoom) const
   {
      Conversion->ZoomBy(Center, Zoom);;
   }

//======================================================================================================================

TWCanvasControllerPrivate::TWCanvasControllerPrivate(TConversion *Conv): Conversion(Conv), IgnoreScrollSignals(false)
   {
      Canvas = INSTANCE(TCanvas);
      Canvas->SetConversion(Conv);
      Viewport = new TViewportWidget(Canvas, Conv);
   }
TWCanvasControllerPrivate::~TWCanvasControllerPrivate()    {   }
void TWCanvasControllerPrivate::ResetScrollBars()
   {
      Q_Q(TWCanvasController);
      QScrollBar *HBar = q->horizontalScrollBar();
      QScrollBar *VBar = q->verticalScrollBar();

      const QRect AllArea = Conversion->GlobalBoundingRect().toRect();

      const int MinX = qMin(AllArea.left(), 0);
      const int MaxX = qMax(AllArea.right(), Viewport->width());
      const int MinY = qMin(AllArea.top(), 0);
      const int MaxY = qMax(AllArea.bottom(), Viewport->height());
      const int dx = MaxX - MinX;
      const int dy = MaxY - MinY;

      bool OldIgnoreScrollSignals = IgnoreScrollSignals;
      IgnoreScrollSignals = true;
      HBar->setRange(0, dx - Viewport->width());
      VBar->setRange(0, dy - Viewport->height());

      const QPointF ViewPos = Conversion->WidgetToViewRect(QPointF());
      const QPointF BarPos(qMax<qreal>(ViewPos.x(), 0.), qMax<qreal>(ViewPos.y(), 0.));
      HBar->setValue(BarPos.x());
      VBar->setValue(BarPos.y());
      IgnoreScrollSignals = OldIgnoreScrollSignals;
   }

void TWCanvasControllerPrivate::ScaleBackgroundToNewSize(Qt::AspectRatioMode AspectRatioMode)
   {
      TShapeList List = INSTANCE(ShapeManager)->FindAll(GraphEd::BackgroundShapeName());
      if (!List.size()) return;
      QImage *ImgPtr = dynamic_cast<QImage *>(List.front().data());
      if (!ImgPtr) return;

      (*ImgPtr) = BackGround.scaled(Conversion->ViewRect().size().toSize(), AspectRatioMode, Qt::SmoothTransformation);
   }

//======================================================================================================================

TWCanvasController::TWCanvasController(TConversion *c): d_ptr(new TWCanvasControllerPrivate(c))
   {
      Q_D(TWCanvasController);
      d->q_ptr = this;
      connect(d->Conversion, SIGNAL(GlobalBoundingRectChanged(QRectF)), this, SLOT(on_GlobalBoundingRectChanged(QRectF)));


      horizontalScrollBar()->setSingleStep(15);
      verticalScrollBar()->setSingleStep(50);
      setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
      setVerticalScrollBarPolicy(Qt::ScrollBarAlwaysOn);
      setViewport(d->Viewport);
      setFrameStyle(QFrame::NoFrame);

      //Set default value
      SetDocumentRect(QRectF(0/*m*/, 0/*m*/, 1/*m*/, 1/*m*/));
   }
void TWCanvasController::resizeEvent(QResizeEvent * /*e*/)
   {
      Q_D(TWCanvasController);
      d->Viewport->ResetLayout();

      horizontalScrollBar()->setPageStep(d->Viewport->width());
      verticalScrollBar()->setPageStep(d->Viewport->height());
      d->ResetScrollBars();
   }
TCanvas *TWCanvasController::GetCanvas() const
   {
      Q_D(const TWCanvasController);
      return d->Canvas;
   }
void TWCanvasController::ZoomIn(const QPointF &WidgetPos)  { ZoomBy(WidgetPos, sqrt(2.0));  }
void TWCanvasController::ZoomOut(const QPointF &WidgetPos) { ZoomBy(WidgetPos, sqrt(0.5));  }
void TWCanvasController::ZoomBy(const QPointF &WidgetPos, qreal Zoom)
   {
      Q_D(TWCanvasController);
      d->Viewport->ZoomBy(WidgetPos, Zoom);
      d->ResetScrollBars();
   }
void TWCanvasController::scrollContentsBy(int dx, int dy)
   {
      Q_D(TWCanvasController);
      if (d->IgnoreScrollSignals) return;
      d->Conversion->SetOffset(d->Conversion->Offset() + QPointF(dx, dy));
      d->Canvas->update();
      emit Scrolled();
   }
void TWCanvasController::wheelEvent(QWheelEvent *e)
   {
      Q_D(TWCanvasController);
      //if (e->modifiers() & Qt::ControlModifier)
         {
            e->delta() > 0 ? ZoomIn(e->pos()) : ZoomOut(e->pos());
            e->accept();
            d->Canvas->update();
            emit Zoomed();
         }
      //else QAbstractScrollArea::wheelEvent(e);
   }
void TWCanvasController::on_GlobalBoundingRectChanged(QRectF New)
   {
      Q_UNUSED(New);
      Q_D(TWCanvasController);
      d->ResetScrollBars();
   }
void TWCanvasController::SetBackground(const QImage &Img, Qt::AspectRatioMode AspectRatioMode)
   {
      if (Img.isNull()) return;
      Q_D(TWCanvasController);
      TImageShapeSP Shape = TImageShapeSP(new TImageShape);
      QImage *Image = dynamic_cast<QImage *>(Shape.data());
      if (!Image || !d->Conversion) return;

      INSTANCE(ShapeManager)->RemoveAll(GraphEd::BackgroundShapeName());

      d->BackGround = Img;
      *Image = Img.scaled(d->Conversion->VirginViewRect().size().toSize(), AspectRatioMode, Qt::SmoothTransformation);
      Shape->SetZIndex(GraphEd::BackgroundIndex);
      Shape->setObjectName(GraphEd::BackgroundShapeName());
      Shape->SetSelectable(false);

      INSTANCE(ShapeManager)->AddShape(Shape);
      INSTANCE(TCanvas)->update();
   }
void TWCanvasController::SetDocumentRect(const QRectF &Rect, Qt::AspectRatioMode AspectRatioMode)
   {
      if (Rect.isEmpty()) return;
      Q_D(TWCanvasController);

      qreal zoom = d->Conversion->Zoom();

      d->Conversion->SetZoom(1);
      d->Conversion->SetDocumentRect(Rect);

      QSizeF NewSize = TMisc::CalcNewSize(d->Viewport->size() - QSize(verticalScrollBar()->width() + d->Conversion->Margin(), horizontalScrollBar()->height() + d->Conversion->Margin()), Rect, TMisc::TSTC_MinSize);

      d->Conversion->SetViewRect(QRectF(QPointF(), NewSize));
      d->ScaleBackgroundToNewSize(AspectRatioMode);

      d->Conversion->ZoomBy(QPointF(NewSize.width()/2, NewSize.height()/2), zoom);

      if (INSTANCE(TWView))
         INSTANCE(TWView)->update_ruler();

      d->ResetScrollBars();
      d->Canvas->repaint();
   }

}
