#include "TWView.h"
#include "core/core.h"
#include "shape/ShapeManager.h"
#include "widgets/canvas/TConversion.h"
#include "widgets/canvas/TCanvas.h"
#include "widgets/canvas/TRuler.h"
#include "tools/ToolManager.h"

#include <QDebug>
#include <QHBoxLayout>
#include <TGraphEditor.h>
#include <widgets/TWToolBox.h>

namespace graphed {

class TWViewPrivate
   {
      Q_DECLARE_PUBLIC(TWView);

   public:
      TWView *q_ptr;
      TWViewPrivate();
      virtual ~TWViewPrivate() { }
      void _q_UpdateBorderDist();
      void _q_UpdateRulerRange();
      void _q_UpdateRuler();

      TRuler *HorizRuler;
      TRuler *VertRuler;
      TWCanvasController *CanvasController;
      TConversion *Conversion;

      bool FirstShowEvent;
   };

TWViewPrivate::TWViewPrivate(): FirstShowEvent(true)
   {
      Conversion = new TConversion;
      CanvasController = new TWCanvasController(Conversion);

      HorizRuler = new TRuler(QwtScaleDraw::BottomScale);
      VertRuler = new TRuler(QwtScaleDraw::LeftScale);
   }
void TWViewPrivate::_q_UpdateBorderDist()
   {
      QPoint TopLeft = (Conversion->ViewRect().topLeft() + QPointF(Conversion->Margin(), Conversion->Margin())).toPoint();
      TopLeft.rx() = qBound(0, TopLeft.x(), HorizRuler->width());
      TopLeft.ry() = qBound(0, TopLeft.y(), VertRuler->height() - HorizRuler->height());

      QPoint BottomRight = (Conversion->ViewRect().bottomRight() + QPointF(Conversion->Margin(), Conversion->Margin())).toPoint();
      BottomRight.rx() = qBound(0, BottomRight.x(), HorizRuler->width());
      BottomRight.ry() = qBound(0, BottomRight.y(), VertRuler->height() - HorizRuler->height());

      //Horiz ruler
      const int LeftDist = TopLeft.x();
      const int RightDist = HorizRuler->width() - BottomRight.x();
      HorizRuler->setBorderDist(LeftDist, RightDist);

      //Vert ruler
      const int UpperDist = TopLeft.y();
      const int BottomDist = VertRuler->height() - BottomRight.y(); // + HorizRuler->height()
      VertRuler->setBorderDist(UpperDist, BottomDist);
   }
void TWViewPrivate::_q_UpdateRulerRange()
   {
      QPointF TopLeft = Conversion->ViewRect().topLeft();
      TopLeft.rx() = qBound<qreal>(0., TopLeft.x(), static_cast<qreal>(INSTANCE(TCanvas)->width()));
      TopLeft.ry() = qBound<qreal>(0., TopLeft.y(), static_cast<qreal>(INSTANCE(TCanvas)->height()));

      QPointF BottomRight = Conversion->ViewRect().bottomRight();
      BottomRight.rx() = qBound<qreal>(0., BottomRight.x(), static_cast<qreal>(INSTANCE(TCanvas)->width()));
      BottomRight.ry() = qBound<qreal>(0., BottomRight.y(), static_cast<qreal>(INSTANCE(TCanvas)->height()));

      TopLeft = Conversion->ViewToPhysical(TopLeft);
      BottomRight = Conversion->ViewToPhysical(BottomRight);

      HorizRuler->setRange(TopLeft.x(), BottomRight.x());
      VertRuler->setRange(BottomRight.y(), TopLeft.y());
   }
void TWViewPrivate::_q_UpdateRuler()
   {
      _q_UpdateBorderDist();
      _q_UpdateRulerRange();
   }

//====================================================================================================================


TWView *TWView::Self = NULL;
bool TWView::IAmBeingCreatedNow = false;
TWView::TWView(): d_ptr(new TWViewPrivate)
   {
      Q_D(TWView);
      d->q_ptr = this;

      QVBoxLayout *VBLRight = new QVBoxLayout;
      VBLRight->setSpacing(0);
      VBLRight->setContentsMargins(0, 0, 0, 0);

      QHBoxLayout *MainLayout = new QHBoxLayout(this);
      MainLayout->setSpacing(0);
      MainLayout->addWidget(d->VertRuler);
      MainLayout->addLayout(VBLRight);

      VBLRight->addWidget(d->CanvasController);
      VBLRight->addWidget(d->HorizRuler);
      d->HorizRuler->setContentsMargins(0, 0, 0, 0);

      setMinimumSize(500, 500);

      connect(d->CanvasController, SIGNAL(Scrolled()), this, SLOT(_q_UpdateRuler()));
      connect(d->CanvasController, SIGNAL(Zoomed()), this, SLOT(_q_UpdateRuler()));
   }
TWView::~TWView()
   {
      delete d_ptr;
      Self = NULL;
   }
TWView *TWView::Instance(const int /*Line*/, const char * /*File*/)
   {
      if (IAmBeingCreatedNow)
         { return NULL; }
      if (!Self)
         {
            IAmBeingCreatedNow = true;
            Self = new TWView;
            IAmBeingCreatedNow = false;
         }
      return Self;
   }
void TWView::Destroy()  { delete Self; Self = 0;  }
void TWView::SetLocked(const bool Lock)
   {
      Q_D(TWView);
      d->CanvasController->GetCanvas()->SetLocked(Lock);
   }
TConversion *TWView::Conversion() const
   {
      Q_D(const TWView);
      return d->Conversion;
   }
TWCanvasController *TWView::CanvasController() const
   {
      Q_D(const TWView);
      return d->CanvasController;
   }
void TWView::update_ruler()
{
    Q_D(TWView);
    d->_q_UpdateRuler();
}

void TWView::resizeEvent(QResizeEvent * /*e*/)
   {
      //QWidget::resizeEvent(e);
      Q_D(TWView);
      d->_q_UpdateRuler();
   }
void TWView::showEvent(QShowEvent * /*e*/)
   {
      Q_D(TWView);
      if (d->FirstShowEvent)
         {
            d->FirstShowEvent = false;
            SetDocumentRect(d->Conversion->DocumentRect());
         }
   }
void TWView::SetDocumentRect(const QRectF &Rect)
   {
      Q_D(TWView);
      d->CanvasController->SetDocumentRect(Rect);
      d->_q_UpdateRuler();

      ((TGraphEditor *)parentWidget())->ToolBox->is_rect_set = false;
   }
void TWView::SetDocumentResolution(const QSize &s)
   {
      Q_D(TWView);
      d->Conversion->SetDocumentResolution(s);
      d->_q_UpdateRuler();
   }

}

#include "moc_TWView.cpp"
