#include "src/modules.hpp"
#include "../moduleplugin.hpp"
#include <QDebug>
#include "modules/eg3000module.h"
#include "modules/oscilloscopemodule.h"
#include "modules/scanmodule.h"
#include "dacscontrol.h"

ModuleManager *ModuleManager::instance = nullptr; //local for module
static
void (ModuleManager::*funload)(I_MODULE *) = nullptr;

I_MODULE::~I_MODULE()
{
    if (ModuleManager::instance)
        //ModuleManager::instance->unload(this);
        if (funload)
            (ModuleManager::instance->*funload)(this);
}


#include <QSettings>
#include "hardware/EG3000.h"
#include "src/scanplot.h"
#include "src/oscilloscope.h"
#include "xyzcalibrations.hpp"
#include "grapheditor/TGraphEditor.h"
#include <grapheditor/widgets/TCheckableButton.h>
#include "grapheditor/widgets/TWMisc.h"
#include <QTimer>

class LithoModule : public I_GUI_MODULE<graphed::TGraphEditor>, public IProcessPacket, public IWriter
{
    Q_OBJECT
public:
    GUI_MODULE_CTOR(LithoModule)
    {
        connect(BStartStop, SIGNAL(BChecked(bool)), this, SLOT(slLitoStartStop(bool)));
        connect(TxyUpdate, SIGNAL(timeout()), this, SLOT(sl_get_curr_xy()));

        QSettings settings("Param/lithography.conf", QSettings::IniFormat);
        BlithoVout->setCurrentIndex(settings.value("VoutType", 0).toInt());
        BScanner->setCurrentIndex(settings.value("ScannerType", 0).toInt());
        BLithoType->setCurrentIndex(settings.value("LithoType", 0).toInt());
        WProperties->getSpeedSlider()->setValue(settings.value("ScanSpeed", 10).toDouble());
    }
    static const QLatin1String name() { return QLatin1String("litho"); }
    void connected(const QLatin1String name, I_MODULE *module) override;

    QPushButton *createPanelButton() override { return new QPushButton(makeHugeIcon(":/MainToolbar/litho.png"), tr("Lithography")); }

    bool processPacket(const BYTE addr, const BYTE sub_addr, const void *data, const DWORD size) override;
    bool sendDataIndirect(IWriter::IEG3KWrite *const writer) override;
    void resetWriter() override;

    graphed::out::TResultsList::const_iterator ResIt;
    graphed::STLPointFVector::const_iterator PointIt;
    EG_LithoParams2 LithoPar;
    TMessageFlag sendStart, sendStop;
    TMessageFlag doGetXY;

public slots:
    void slLitoStartStop(bool sig);
    void sl_LithoNextpacket();
    void sl_get_curr_xy();
};

class LithoPlugin : public QObject, public IModulePlugin
{
    Q_OBJECT
    Q_INTERFACES(IModulePlugin)
#if (QT_VERSION >= QT_VERSION_CHECK(5, 0, 0))
    Q_PLUGIN_METADATA(IID "imp")
#endif
public:
    void init(ModuleManager *mgr, void (ModuleManager::*_funload)(I_MODULE *)) override
    {
        ModuleManager::instance = mgr;
        funload = _funload;
        ModuleManager::instance->addToFactoryGui<LithoModule>();
    }
    I_MODULE *getNew(QWidget *parent = nullptr) override
    {
        return new LithoModule(parent);
    }
};
#if (QT_VERSION < QT_VERSION_CHECK(5, 0, 0))
Q_EXPORT_PLUGIN2(litho, LithoPlugin)
#endif

//---------------------------------------------------------------------------------------

#include <grapheditor/widgets/canvas/TWView.h>
void LithoModule::connected(const QLatin1String name, I_MODULE *module)
{
    if (name == EG3000Module::name()) {
        EG3000_CONTROL *m = (EG3000_CONTROL *)module->basecast();

        m->rThread.readers.add(this);
        m->wThread.writers.add(this);
    }
    if (name == ScanModule::name()) {
        ScanPlot *m = (ScanPlot *)module->basecast();
        connect(m, SIGNAL(SendRect(QRectF)), GetView(), SLOT(SetDocumentRect(QRectF)));
        connect(m, SIGNAL(SendBackground(QImage)), GetView()->CanvasController(), SLOT(SetBackground(QImage)));

        m->findChild<QPushButton *>("BSendToGraphEditor")->show();
    }
    if (name == OscilloscopeModule::name()) {
        Oscilloscope *m = (Oscilloscope *)module->basecast();

        //soft xy sync
        connect(this, SIGNAL(syncX1pos(double)), m->DacsWidget, SIGNAL(xUpdate(double)));
        connect(this, SIGNAL(syncY1pos(double)), m->DacsWidget, SIGNAL(yUpdate(double)));
        connect(this, SIGNAL(syncX2pos(double)), m->DacsWidget, SIGNAL(x2Update(double)));
        connect(this, SIGNAL(syncY2pos(double)), m->DacsWidget, SIGNAL(y2Update(double)));

        connect(m->DacsWidget, SIGNAL(xUpdate(double)), this, SLOT(setX1(double)));
        connect(m->DacsWidget, SIGNAL(x2Update(double)), this, SLOT(setX2(double)));
        connect(m->DacsWidget, SIGNAL(yUpdate(double)), this, SLOT(setY1(double)));
        connect(m->DacsWidget, SIGNAL(y2Update(double)), this, SLOT(setY2(double)));
        connect(m->DacsWidget, SIGNAL(zUpdate(double)), this, SLOT(updZ1(double)));
        connect(m->DacsWidget, SIGNAL(z2Update(double)), this, SLOT(updZ2(double)));

        connect(this, SIGNAL(sig_XYPosition(double, double)), m->DacsWidget, SLOT(Go_to_XY(double, double)));
        connect(this, SIGNAL(sig_XYPosition2(double, double)), m->DacsWidget, SLOT(Go_to_XY2(double, double)));
        connect(this, SIGNAL(sig_ZPosition(double)), m->DacsWidget, SLOT(Go_to_Z(double)));
        connect(this, SIGNAL(sig_ZPosition2(double)), m->DacsWidget, SLOT(Go_to_Z2(double)));

        connect(m->CalibWidget, SIGNAL(sig_XYZ_range(double,double,double)), this, SLOT(sl_setHeadRange(double, double, double)));
        connect(m->CalibWidget, SIGNAL(sig_XYZ_range2(double,double,double)), this, SLOT(sl_setStageRange(double, double, double)));
    }
}
bool LithoModule::processPacket(const BYTE addr, const BYTE sub_addr, const void *data, const DWORD )
{
    if (addr != ADR_CPU)
        return false;

    if (sub_addr == CPU_LITHO_DONE) {
        QMetaObject::invokeMethod(this, "sl_LithoNextpacket");
        QMetaObject::invokeMethod(this, "updatePrbar");

        return true;
    }

    if (sub_addr == CPU_GET_DATA) {
        EG_GetData *Get_data = (EG_GetData *)data;
        BYTE src = Get_data->Source;
        DWORD data = Get_data->Data;

        if (src == srcDACX)
            QMetaObject::invokeMethod(this, "setX1", Q_ARG(double, data / 65535. * xHeadRange));
        if (src == srcDACY)
            QMetaObject::invokeMethod(this, "setY1", Q_ARG(double, data / 65535. * yHeadRange));
        if (src == srcDACX2)
            QMetaObject::invokeMethod(this, "setX2", Q_ARG(double, data / 65535. * xStageRange));
        if (src == srcDACY2)
            QMetaObject::invokeMethod(this, "setY2", Q_ARG(double, data / 65535. * yStageRange));

        return false;
    }

    return false;
}

#include "grapheditor/tools/TApplyTool.h"
#include "grapheditor/tools/ToolManager.h"
#include <QCheckBox>
void LithoModule::slLitoStartStop(bool sig)
{
    if (!sig) {
        sendStop = true;
        return;
    }

    LithoPar.LithoMode = BLithoType->currentIndex()+1;
    LithoPar.Scanner = BScanner->currentIndex();
    LithoPar.isContinuous = CBisContinuous->isChecked();
    LithoPar.DACIndex = BlithoVout->currentIndex();

    graphed::TApplyTool *ApplyTool = (graphed::TApplyTool *)FindTool("TApplyTool");
    ResIt = ApplyTool->ResultsList().begin();
    if (ResIt != ApplyTool->ResultsList().end())
        PointIt = (**ResIt).PointList().begin();

    sl_LithoNextpacket();
}

void LithoModule::sl_LithoNextpacket()
{
    graphed::TApplyTool *ApplyTool = (graphed::TApplyTool *)FindTool("TApplyTool");
    if (ResIt == ApplyTool->ResultsList().end()) {
        Finish();
        return;
    }

    graphed::out::TResults &Result = **ResIt;
    LithoPar.Speed = Result.Speed();
    LithoPar.V = Result.VtData1(PointIt).first;
    LithoPar.T = Result.VtData1(PointIt).second;
    if (LithoPar.LithoMode == ltmSync)
        LithoPar.V = 1;

    LithoPar.Size = 0;
    while (PointIt != Result.PointList().end() && LithoPar.Size < LITHO_DATA_SIZE) {
        if (LithoPar.LithoMode != ltmSync &&
                fabs(Result.VtData1(PointIt).first - LithoPar.V) >= 1e-5)
            break;

        QPointF fl1 = *PointIt;
        LithoPar.XY[LithoPar.Size*2] = 1e6 * fl1.x();
        LithoPar.XY[LithoPar.Size*2+1] = 1e6 * fl1.y();
        LithoPar.Size++;
        PointIt++;
    }

    if (PointIt == Result.PointList().end()) {
        ResIt++;
        if (ResIt != ApplyTool->ResultsList().end())
            PointIt = (**ResIt).PointList().begin();
    }

    sendStart = true;
}
#include <defs.h>
bool LithoModule::sendDataIndirect(IWriter::IEG3KWrite *const writer)
{
    bool wrote = false;

    if (sendStart.claimMessage()) {
        wrote = true;
        writer->send_packet(ADR_CPU, CPU_LITHO_START2, &LithoPar, 5*4 + LithoPar.Size*2*4); //sizeof(EG_LithoParams2));
    }
    if (sendStop.claimMessage()) {
        wrote = true;
        writer->send_empty_packet(ADR_CPU, CPU_LITHO_STOP);
    }
    if (doGetXY.claimMessage()) {
        wrote = true;
        writer->send_word_packet(ADR_CPU, CPU_GET_DATA, srcDACX);
        writer->send_word_packet(ADR_CPU, CPU_GET_DATA, srcDACY);
        writer->send_word_packet(ADR_CPU, CPU_GET_DATA, srcDACX2);
        writer->send_word_packet(ADR_CPU, CPU_GET_DATA, srcDACY2);
    }

    return wrote;
}
void LithoModule::resetWriter()
{
    sendStart = false;
    sendStop = false;
    doGetXY = false;
}
void LithoModule::sl_get_curr_xy()
{
    doGetXY = true;
}


#include <litho.moc>
